"""Tests for ggsci.scales (discrete classes, continuous functions, aliases)."""

from __future__ import annotations

import inspect
from collections.abc import Callable

import pytest
from plotnine.scales import scale_color_gradientn, scale_fill_gradientn

import ggsci.scales as sc


def _discrete_scale_classes() -> list[tuple[str, type]]:
    items: list[tuple[str, type]] = []
    for name, obj in sc.__dict__.items():
        if not isinstance(obj, type):
            continue
        if not issubclass(obj, sc.scale_discrete):
            continue
        # Skip the base class itself
        if name == "scale_discrete":
            continue
        items.append((name, obj))
    return items


def _continuous_scale_factories() -> list[tuple[str, Callable[..., object]]]:
    return [
        (name, obj)
        for name, obj in sc.__dict__.items()
        if name.startswith("scale_color_")
        and callable(obj)
        and not isinstance(obj, type)
        and "gradientn" in inspect.getsource(sc.__dict__[name]).lower()
    ] + [
        (name, obj)
        for name, obj in sc.__dict__.items()
        if name.startswith("scale_fill_")
        and callable(obj)
        and not isinstance(obj, type)
        and "gradientn" in inspect.getsource(sc.__dict__[name]).lower()
    ]


@pytest.mark.parametrize("name,cls", _discrete_scale_classes())
def test_discrete_scales_aesthetics_and_palette(name: str, cls: type):
    s = cls()
    if name.startswith("scale_color_") or name.startswith("scale_colour_"):
        assert s._aesthetics == ["color"]
    elif name.startswith("scale_fill_"):
        assert s._aesthetics == ["fill"]
    else:
        pytest.fail(f"Unexpected discrete scale name: {name}")

    # Palette is a callable taking n and returning list[str]
    colors = s.palette(3)
    assert isinstance(colors, list) and len(colors) == 3
    assert all(c.startswith("#") and len(c) == 7 for c in colors)

    # Alpha is applied via InitVar
    s_alpha = cls(alpha=0.6)
    out = s_alpha.palette(1)
    assert len(out) == 1 and out[0].startswith("#") and len(out[0]) == 9
    assert out[0][-2:] == f"{int(0.6 * 255):02x}"


@pytest.mark.parametrize("name,fn", _continuous_scale_factories())
def test_continuous_scale_return_types(name: str, fn: Callable[..., object]):
    obj = fn()
    if name.startswith("scale_color_"):
        assert isinstance(obj, scale_color_gradientn)
    else:
        assert isinstance(obj, scale_fill_gradientn)


def test_british_aliases_identity():
    # All scale_colour_* should be the same object as scale_color_*
    uk_names = [n for n in sc.__dict__ if n.startswith("scale_colour_")]
    for uk in uk_names:
        us = uk.replace("colour", "color")
        assert hasattr(sc, us)
        assert getattr(sc, uk) is getattr(sc, us)


def test_init_exports_alignment():
    # Importing from package root should expose the same objects
    import ggsci as pkg

    names = [
        # Sample a few across types to ensure import surface
        "scale_color_npg",
        "scale_fill_npg",
        "scale_colour_npg",
        "scale_color_gsea",
        "scale_fill_bs5",
        "pal_npg",
        "pal_gsea",
    ]
    for name in names:
        assert hasattr(pkg, name)
        # Identity with module definitions
        mod = (
            sc
            if name.startswith("scale_")
            else __import__("ggsci.palettes", fromlist=[name])
        )
        assert getattr(pkg, name) is getattr(mod, name)



"""Tests for ggsci.palette functions (discrete and continuous)."""

from __future__ import annotations

import inspect
from collections.abc import Callable

import pytest

from ggsci import palettes as pl
from ggsci.data import PALETTES
from ggsci.data_iterm import ITERM_PALETTES, PALETTES_ITERM


def _is_continuous_palette(func: Callable[..., object]) -> bool:
    sig = inspect.signature(func)
    return "n" in sig.parameters


def _alpha_hex(alpha: float) -> str:
    return f"{int(alpha * 255):02x}"


def _all_palette_functions() -> dict[str, Callable[..., object]]:
    return {name[4:]: getattr(pl, name) for name in dir(pl) if name.startswith("pal_")}


def _discrete_and_continuous() -> tuple[list[str], list[str]]:
    funcs = _all_palette_functions()
    discrete = [name for name, fn in funcs.items() if not _is_continuous_palette(fn)]
    continuous = [name for name, fn in funcs.items() if _is_continuous_palette(fn)]
    return sorted(discrete), sorted(continuous)


DISCRETE_NAMES, CONTINUOUS_NAMES = _discrete_and_continuous()


def _iter_discrete_palette_cases(
    name: str,
) -> list[tuple[str, list[str], dict[str, str]]]:
    if name == "iterm":
        cases: list[tuple[str, list[str], dict[str, str]]] = []
        for palette_key, variants in PALETTES_ITERM.items():
            for variant, colors in variants.items():
                cases.append((palette_key, list(colors), {"variant": variant}))
        return cases

    return [
        (palette_key, list(colors), {})
        for palette_key, colors in PALETTES[name].items()
    ]


@pytest.mark.parametrize("name", DISCRETE_NAMES)
def test_discrete_palette_happy_path_and_alpha(name: str):
    func: Callable[..., Callable[[int], list[str]]] = getattr(pl, f"pal_{name}")

    # Validate each available sub-palette for this family
    palette_iter = _iter_discrete_palette_cases(name)

    for palette_key, colors, extra_kwargs in palette_iter:
        # Happy path
        pal_fn = func(palette=palette_key, alpha=1.0, **extra_kwargs)
        n = min(3, len(colors))
        out = pal_fn(n)
        assert isinstance(out, list) and len(out) == n
        assert all(c.startswith("#") and len(c) == 7 for c in out)

        # Too many requested colors -> error
        with pytest.raises(ValueError):
            pal_fn(len(colors) + 1)

        # Alpha applied in discrete palette function
        pal_fn_a = func(palette=palette_key, alpha=0.6, **extra_kwargs)
        out_a = pal_fn_a(1)
        assert len(out_a) == 1 and out_a[0].startswith("#") and len(out_a[0]) == 9
        assert out_a[0][-2:] == _alpha_hex(0.6)


@pytest.mark.parametrize("name", DISCRETE_NAMES)
def test_discrete_palette_errors(name: str):
    func: Callable[..., Callable[[int], list[str]]] = getattr(pl, f"pal_{name}")

    with pytest.raises(ValueError):
        func(palette="__unknown__", alpha=1.0)

    if name == "iterm":
        first_palette = ITERM_PALETTES[0]
        with pytest.raises(ValueError):
            func(palette=first_palette, variant="__unknown__")

    for bad_alpha in (0.0, -0.1, 1.0 + 1e-9):
        with pytest.raises(ValueError):
            func(alpha=bad_alpha)


@pytest.mark.parametrize("name", CONTINUOUS_NAMES)
def test_continuous_palette_happy_path_reverse_alpha(name: str):
    func: Callable[..., list[str]] = getattr(pl, f"pal_{name}")

    # Exercise all palettes for the family (kept small n for speed)
    for palette_key in PALETTES[name].keys():
        # Forward
        out = func(palette=palette_key, n=7, alpha=1.0, reverse=False)
        assert isinstance(out, list) and len(out) == 7
        assert all(c.startswith("#") and len(c) == 7 for c in out)

        # Reverse
        out_r = func(palette=palette_key, n=7, alpha=1.0, reverse=True)
        assert out_r == out[::-1]

        # Alpha applied post-interpolation
        out_a = func(palette=palette_key, n=5, alpha=0.6, reverse=False)
        assert all(c.startswith("#") and len(c) == 9 for c in out_a)
        assert out_a[0][-2:] == _alpha_hex(0.6)


@pytest.mark.parametrize("name", CONTINUOUS_NAMES)
def test_continuous_palette_errors(name: str):
    func: Callable[..., list[str]] = getattr(pl, f"pal_{name}")

    with pytest.raises(ValueError):
        func(palette="__unknown__")

    for bad_alpha in (0.0, -0.1, 1.0 + 1e-9):
        with pytest.raises(ValueError):
            func(alpha=bad_alpha)



"""
Scientific journal and sci-fi themed color palettes for plotnine.
"""

from .data_iterm import ITERM_PALETTES
from .examples import example_barplot, example_scatterplot
from .palettes import (
    pal_aaas,
    pal_atlassian,
    pal_bmj,
    pal_bs5,
    pal_cosmic,
    pal_d3,
    pal_flatui,
    pal_frontiers,
    pal_futurama,
    pal_gsea,
    pal_igv,
    pal_iterm,
    pal_jama,
    pal_jco,
    pal_lancet,
    pal_locuszoom,
    pal_material,
    pal_nejm,
    pal_npg,
    pal_observable,
    pal_primer,
    pal_rickandmorty,
    pal_simpsons,
    pal_startrek,
    pal_tron,
    pal_tw3,
    pal_uchicago,
    pal_ucscgb,
)
from .scales import (
    scale_color_aaas,
    scale_color_atlassian,
    scale_color_bmj,
    scale_color_bs5,
    scale_color_cosmic,
    scale_color_d3,
    scale_color_flatui,
    scale_color_frontiers,
    scale_color_futurama,
    scale_color_gsea,
    scale_color_igv,
    scale_color_iterm,
    scale_color_jama,
    scale_color_jco,
    scale_color_lancet,
    scale_color_locuszoom,
    scale_color_material,
    scale_color_nejm,
    scale_color_npg,
    scale_color_observable,
    scale_color_primer,
    scale_color_rickandmorty,
    scale_color_simpsons,
    scale_color_startrek,
    scale_color_tron,
    scale_color_tw3,
    scale_color_uchicago,
    scale_color_ucscgb,
    scale_colour_aaas,
    scale_colour_atlassian,
    scale_colour_bmj,
    scale_colour_bs5,
    scale_colour_cosmic,
    scale_colour_d3,
    scale_colour_flatui,
    scale_colour_frontiers,
    scale_colour_futurama,
    scale_colour_gsea,
    scale_colour_igv,
    scale_colour_iterm,
    scale_colour_jama,
    scale_colour_jco,
    scale_colour_lancet,
    scale_colour_locuszoom,
    scale_colour_material,
    scale_colour_nejm,
    scale_colour_npg,
    scale_colour_observable,
    scale_colour_primer,
    scale_colour_rickandmorty,
    scale_colour_simpsons,
    scale_colour_startrek,
    scale_colour_tron,
    scale_colour_tw3,
    scale_colour_uchicago,
    scale_colour_ucscgb,
    scale_fill_aaas,
    scale_fill_atlassian,
    scale_fill_bmj,
    scale_fill_bs5,
    scale_fill_cosmic,
    scale_fill_d3,
    scale_fill_flatui,
    scale_fill_frontiers,
    scale_fill_futurama,
    scale_fill_gsea,
    scale_fill_igv,
    scale_fill_iterm,
    scale_fill_jama,
    scale_fill_jco,
    scale_fill_lancet,
    scale_fill_locuszoom,
    scale_fill_material,
    scale_fill_nejm,
    scale_fill_npg,
    scale_fill_observable,
    scale_fill_primer,
    scale_fill_rickandmorty,
    scale_fill_simpsons,
    scale_fill_startrek,
    scale_fill_tron,
    scale_fill_tw3,
    scale_fill_uchicago,
    scale_fill_ucscgb,
)



"""Tests for ggsci.utils helpers."""

from __future__ import annotations

import pytest

from ggsci.utils import apply_alpha, hex_to_rgb, interpolate_colors, rgb_to_hex


def test_hex_to_rgb_and_back():
    # Basic round-trip for full intensity channels
    assert hex_to_rgb("#ffffff") == (255, 255, 255)
    assert hex_to_rgb("#000000") == (0, 0, 0)

    # to_hex returns lowercase, verify known mapping
    assert rgb_to_hex((1.0, 0.0, 0.0)) == "#ff0000"
    assert rgb_to_hex((0.0, 1.0, 0.0)) == "#00ff00"
    assert rgb_to_hex((0.0, 0.0, 1.0)) == "#0000ff"


@pytest.mark.parametrize(
    "colors,alpha,expected_suffix",
    [
        (
            [
                "#ffffff",
            ],
            0.5,
            "7f",
        ),
        (["#000000"], 1.0, "ff"),
        (["#abcdef"], 0.0 + 1e-9, "00"),
    ],
)
def test_apply_alpha(colors: list[str], alpha: float, expected_suffix: str):
    out = apply_alpha(colors, alpha)
    assert len(out) == len(colors)
    assert all(v.startswith("#") and len(v) == 9 for v in out)
    # Last two digits encode alpha
    assert out[0][-2:] == expected_suffix


def test_interpolate_colors_endpoints_and_sampling():
    # Interpolation across two endpoints keeps ends intact
    colors = interpolate_colors(["#ff0000", "#0000ff"], 5)
    assert len(colors) == 5
    assert colors[0] == "#ff0000"
    assert colors[-1] == "#0000ff"

    # When n <= len(colors), sample endpoints evenly
    base = ["#000000", "#111111", "#222222"]
    sampled = interpolate_colors(base, 2)
    assert sampled == [base[0], base[-1]]



"""
Utility functions for color palette manipulation.
"""

from collections.abc import Sequence

import numpy as np
from matplotlib.colors import LinearSegmentedColormap, to_hex, to_rgb


def hex_to_rgb(hex_color: str) -> tuple[int, int, int]:
    """
    Convert a hex color to an RGB tuple.

    Args:
        hex_color: Color in hex form (e.g., "#RRGGBB").

    Returns:
        A tuple of three integers in the 0 to 255 range.
    """
    rgb = to_rgb(hex_color)
    return (
        int(rgb[0] * 255),
        int(rgb[1] * 255),
        int(rgb[2] * 255),
    )


def rgb_to_hex(rgb: tuple[float, float, float]) -> str:
    """
    Convert an RGB tuple to a hex color.

    Args:
        rgb: A tuple of three floats in the 0 to 1 range.

    Returns:
        A hex color string (e.g., "#RRGGBB").
    """
    return to_hex(rgb)


def apply_alpha(colors: Sequence[str], alpha: float) -> Sequence[str]:
    """
    Apply alpha transparency to colors.

    Args:
        colors: Sequence of hex color codes.
        alpha: Alpha value between 0 and 1.

    Returns:
        Colors with alpha applied, as 8-digit RGBA hex strings.
    """
    result: list[str] = []
    for color in colors:
        rgb = to_rgb(color)
        rgba_hex = "#{:02x}{:02x}{:02x}{:02x}".format(
            int(rgb[0] * 255),
            int(rgb[1] * 255),
            int(rgb[2] * 255),
            int(alpha * 255),
        )
        result.append(rgba_hex)
    return result


def interpolate_colors(colors: Sequence[str], n: int) -> Sequence[str]:
    """
    Interpolate between colors to generate a sequence of length n.

    Args:
        colors: Base colors to interpolate between.
        n: Number of colors to generate.

    Returns:
        A list of n interpolated hex colors.
    """
    if n <= len(colors):
        # If requesting fewer colors than available, sample evenly
        indices = np.linspace(0, len(colors) - 1, n).astype(int)
        return [colors[i] for i in indices]

    # Create a colormap from the base colors
    cmap = LinearSegmentedColormap.from_list("custom", list(colors), N=n)

    # Generate n evenly spaced colors
    positions = np.linspace(0, 1, n)
    interpolated: list[str] = []

    for pos in positions:
        rgba = cmap(pos)
        # Convert to hex (ignore alpha channel)
        hex_color = to_hex(rgba[:3])
        interpolated.append(hex_color)

    return interpolated



"""
Palette generators for ggsci.

This module provides palette functions that either return a callable for
discrete palettes or a sequence of colors for continuous palettes.
"""

from collections.abc import Callable, Sequence
from typing import Final, TypeAlias

from .data import PALETTES
from .data_iterm import ITERM_PALETTES, PALETTES_ITERM
from .utils import apply_alpha, interpolate_colors

PaletteFunc: TypeAlias = Callable[[int], Sequence[str]]

ITERM_VARIANTS: Final[tuple[str, ...]] = ("normal", "bright")


def pal_npg(palette: str = "nrc", alpha: float = 1.0) -> PaletteFunc:
    """
    NPG journal color palette.

    Args:
        palette: Palette name. Currently only "nrc" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["npg"]:
        raise ValueError(f"Unknown NPG palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["npg"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_aaas(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    AAAS journal color palette.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["aaas"]:
        raise ValueError(f"Unknown AAAS palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["aaas"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_nejm(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    NEJM journal color palette.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["nejm"]:
        raise ValueError(f"Unknown NEJM palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["nejm"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_lancet(palette: str = "lanonc", alpha: float = 1.0) -> PaletteFunc:
    """
    Lancet journal color palette.

    Args:
        palette: Palette name. Currently only "lanonc" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["lancet"]:
        raise ValueError(f"Unknown Lancet palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["lancet"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_jama(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    JAMA journal color palette.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["jama"]:
        raise ValueError(f"Unknown JAMA palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["jama"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_bmj(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    BMJ journal color palette.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["bmj"]:
        raise ValueError(f"Unknown BMJ palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["bmj"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_jco(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    JCO journal color palette.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["jco"]:
        raise ValueError(f"Unknown JCO palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["jco"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_ucscgb(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    UCSC Genome Browser color palette.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["ucscgb"]:
        raise ValueError(f"Unknown UCSCGB palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["ucscgb"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_d3(palette: str = "category10", alpha: float = 1.0) -> PaletteFunc:
    """
    D3.js color palette.

    Args:
        palette: Palette name: "category10", "category20", "category20b", or "category20c".
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["d3"]:
        raise ValueError(f"Unknown D3 palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["d3"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_observable(palette: str = "observable10", alpha: float = 1.0) -> PaletteFunc:
    """
    Observable color palette.

    Args:
        palette: Palette name. Currently only "observable10" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["observable"]:
        raise ValueError(f"Unknown Observable palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["observable"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_primer(palette: str = "mark17", alpha: float = 1.0) -> PaletteFunc:
    """
    Primer design system color palette.

    Args:
        palette: Palette name. Currently only "mark17" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["primer"]:
        raise ValueError(f"Unknown Primer palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["primer"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_atlassian(palette: str = "categorical8", alpha: float = 1.0) -> PaletteFunc:
    """
    Atlassian design system color palette.

    Args:
        palette: Palette name. Currently only "categorical8" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["atlassian"]:
        raise ValueError(f"Unknown Atlassian palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["atlassian"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_iterm(
    palette: str = "Rose Pine",
    variant: str = "normal",
    alpha: float = 1.0,
) -> PaletteFunc:
    """
    iTerm color palette.

    Args:
        palette: Palette name. See `ITERM_PALETTES` for available options.
        variant: Palette variant. Either "normal" or "bright".
        alpha: Transparency level, between 0 and 1.

    Details:
        Preview all iTerm palettes: <https://nanx.me/ggsci-iterm/>.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name, variant, or alpha is invalid.
    """
    if palette not in ITERM_PALETTES:
        raise ValueError(f"Unknown iTerm palette: {palette}")

    if variant not in ITERM_VARIANTS:
        raise ValueError(f"Unknown iTerm variant: {variant}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES_ITERM[palette][variant]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' ({variant}) has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_locuszoom(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    LocusZoom color palette.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["locuszoom"]:
        raise ValueError(f"Unknown LocusZoom palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["locuszoom"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_igv(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    IGV color palette.

    Args:
        palette: Palette name: "default" or "alternating".
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["igv"]:
        raise ValueError(f"Unknown IGV palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["igv"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_cosmic(palette: str = "hallmarks_dark", alpha: float = 1.0) -> PaletteFunc:
    """
    COSMIC color palette.

    Args:
        palette: Palette name: "hallmarks_dark", "hallmarks_light", or "signature_substitutions".
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["cosmic"]:
        raise ValueError(f"Unknown COSMIC palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["cosmic"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_uchicago(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    University of Chicago color palette.

    Args:
        palette: Palette name: "default", "light", or "dark".
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["uchicago"]:
        raise ValueError(f"Unknown UChicago palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["uchicago"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_startrek(palette: str = "uniform", alpha: float = 1.0) -> PaletteFunc:
    """
    Star Trek color palette.

    Args:
        palette: Palette name. Currently only "uniform" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["startrek"]:
        raise ValueError(f"Unknown Star Trek palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["startrek"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_tron(palette: str = "legacy", alpha: float = 1.0) -> PaletteFunc:
    """
    Tron Legacy color palette.

    Args:
        palette: Palette name. Currently only "legacy" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["tron"]:
        raise ValueError(f"Unknown Tron palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["tron"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_futurama(palette: str = "planetexpress", alpha: float = 1.0) -> PaletteFunc:
    """
    Futurama color palette.

    Args:
        palette: Palette name. Currently only "planetexpress" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["futurama"]:
        raise ValueError(f"Unknown Futurama palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["futurama"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_rickandmorty(palette: str = "schwifty", alpha: float = 1.0) -> PaletteFunc:
    """
    Rick and Morty color palette.

    Args:
        palette: Palette name. Currently only "schwifty" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["rickandmorty"]:
        raise ValueError(f"Unknown Rick and Morty palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["rickandmorty"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_simpsons(palette: str = "springfield", alpha: float = 1.0) -> PaletteFunc:
    """
    The Simpsons color palette.

    Args:
        palette: Palette name. Currently only "springfield" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["simpsons"]:
        raise ValueError(f"Unknown Simpsons palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["simpsons"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_flatui(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    Flat UI color palette.

    Args:
        palette: Palette name: "default", "flattastic", or "aussie".
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["flatui"]:
        raise ValueError(f"Unknown Flat UI palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["flatui"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_frontiers(palette: str = "default", alpha: float = 1.0) -> PaletteFunc:
    """
    Frontiers journal color palette.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.

    Returns:
        A callable that takes n and returns a color sequence.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["frontiers"]:
        raise ValueError(f"Unknown Frontiers palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    colors = PALETTES["frontiers"][palette]

    def palette_func(n: int) -> Sequence[str]:
        if n > len(colors):
            raise ValueError(
                f"Palette '{palette}' has only {len(colors)} colors, "
                f"but {n} were requested"
            )
        selected = colors[:n]
        if alpha < 1:
            return apply_alpha(selected, alpha)
        return selected

    return palette_func


def pal_gsea(
    palette: str = "default",
    n: int = 12,
    alpha: float = 1.0,
    reverse: bool = False,
) -> Sequence[str]:
    """
    GSEA GenePattern color palette (continuous/diverging).

    Args:
        palette: Palette name. Currently only "default" is available.
        n: Number of colors to generate.
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.

    Returns:
        A sequence of hex color codes.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["gsea"]:
        raise ValueError(f"Unknown GSEA palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    base_colors = PALETTES["gsea"][palette]
    colors = interpolate_colors(base_colors, n)

    if reverse:
        colors = colors[::-1]

    if alpha < 1:
        colors = apply_alpha(colors, alpha)

    return colors


def pal_bs5(
    palette: str = "blue",
    n: int = 10,
    alpha: float = 1.0,
    reverse: bool = False,
) -> Sequence[str]:
    """
    Bootstrap 5 color palette (continuous/sequential).

    Args:
        palette: Palette name: "blue", "indigo", "purple", "pink", "red",
            "orange", "yellow", "green", "teal", "cyan", or "gray".
        n: Number of colors to generate.
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.

    Returns:
        A sequence of hex color codes.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["bs5"]:
        raise ValueError(f"Unknown BS5 palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    base_colors = PALETTES["bs5"][palette]
    colors = interpolate_colors(base_colors, n)

    if reverse:
        colors = colors[::-1]

    if alpha < 1:
        colors = apply_alpha(colors, alpha)

    return colors


def pal_material(
    palette: str = "red",
    n: int = 10,
    alpha: float = 1.0,
    reverse: bool = False,
) -> Sequence[str]:
    """
    Material Design color palette (continuous/sequential).

    Args:
        palette: Palette name: "red", "pink", "purple", "deep-purple", "indigo",
            "blue", "light-blue", "cyan", "teal", "green", "light-green",
            "lime", "yellow", "amber", "orange", "deep-orange", "brown",
            "grey", or "blue-grey".
        n: Number of colors to generate.
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.

    Returns:
        A sequence of hex color codes.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["material"]:
        raise ValueError(f"Unknown Material palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    base_colors = PALETTES["material"][palette]
    colors = interpolate_colors(base_colors, n)

    if reverse:
        colors = colors[::-1]

    if alpha < 1:
        colors = apply_alpha(colors, alpha)

    return colors


def pal_tw3(
    palette: str = "blue",
    n: int = 11,
    alpha: float = 1.0,
    reverse: bool = False,
) -> Sequence[str]:
    """
    Tailwind CSS 3 color palette (continuous/sequential).

    Args:
        palette: Palette name: "slate", "gray", "zinc", "neutral", "stone",
            "red", "orange", "amber", "yellow", "lime", "green", "emerald",
            "teal", "cyan", "sky", "blue", "indigo", "violet", "purple",
            "fuchsia", "pink", or "rose".
        n: Number of colors to generate.
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.

    Returns:
        A sequence of hex color codes.

    Raises:
        ValueError: If the palette name is unknown or alpha is invalid.
    """
    if palette not in PALETTES["tw3"]:
        raise ValueError(f"Unknown Tailwind CSS 3 palette: {palette}")

    if not 0 < alpha <= 1:
        raise ValueError("alpha must be in (0, 1]")

    base_colors = PALETTES["tw3"][palette]
    colors = interpolate_colors(base_colors, n)

    if reverse:
        colors = colors[::-1]

    if alpha < 1:
        colors = apply_alpha(colors, alpha)

    return colors



"""
Plotnine scales for ggsci palettes.
"""

from dataclasses import InitVar, dataclass
from typing import Any

from plotnine.scales import scale_color_gradientn, scale_fill_gradientn
from plotnine.scales.scale_discrete import scale_discrete

from .palettes import (
    pal_aaas,
    pal_atlassian,
    pal_bmj,
    pal_bs5,
    pal_cosmic,
    pal_d3,
    pal_flatui,
    pal_frontiers,
    pal_futurama,
    pal_gsea,
    pal_igv,
    pal_iterm,
    pal_jama,
    pal_jco,
    pal_lancet,
    pal_locuszoom,
    pal_material,
    pal_nejm,
    pal_npg,
    pal_observable,
    pal_primer,
    pal_rickandmorty,
    pal_simpsons,
    pal_startrek,
    pal_tron,
    pal_tw3,
    pal_uchicago,
    pal_ucscgb,
)


@dataclass
class scale_color_npg(scale_discrete):
    """
    NPG journal color scale.

    Args:
        palette: Palette name. Currently only "nrc" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "nrc"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_npg(palette, alpha))


@dataclass
class scale_fill_npg(scale_discrete):
    """
    NPG journal fill scale.

    Args:
        palette: Palette name. Currently only "nrc" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "nrc"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_npg(palette, alpha))


@dataclass
class scale_color_aaas(scale_discrete):
    """
    AAAS journal color scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_aaas(palette, alpha))


@dataclass
class scale_fill_aaas(scale_discrete):
    """
    AAAS journal fill scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_aaas(palette, alpha))


@dataclass
class scale_color_nejm(scale_discrete):
    """
    NEJM journal color scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_nejm(palette, alpha))


@dataclass
class scale_fill_nejm(scale_discrete):
    """
    NEJM journal fill scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_nejm(palette, alpha))


@dataclass
class scale_color_lancet(scale_discrete):
    """
    Lancet journal color scale.

    Args:
        palette: Palette name. Currently only "lanonc" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "lanonc"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_lancet(palette, alpha))


@dataclass
class scale_fill_lancet(scale_discrete):
    """
    Lancet journal fill scale.

    Args:
        palette: Palette name. Currently only "lanonc" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "lanonc"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_lancet(palette, alpha))


@dataclass
class scale_color_jama(scale_discrete):
    """
    JAMA journal color scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_jama(palette, alpha))


@dataclass
class scale_fill_jama(scale_discrete):
    """
    JAMA journal fill scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_jama(palette, alpha))


@dataclass
class scale_color_bmj(scale_discrete):
    """
    BMJ journal color scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_bmj(palette, alpha))


@dataclass
class scale_fill_bmj(scale_discrete):
    """
    BMJ journal fill scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_bmj(palette, alpha))


@dataclass
class scale_color_jco(scale_discrete):
    """
    JCO journal color scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_jco(palette, alpha))


@dataclass
class scale_fill_jco(scale_discrete):
    """
    JCO journal fill scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_jco(palette, alpha))


@dataclass
class scale_color_ucscgb(scale_discrete):
    """
    UCSC Genome Browser color scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_ucscgb(palette, alpha))


@dataclass
class scale_fill_ucscgb(scale_discrete):
    """
    UCSC Genome Browser fill scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_ucscgb(palette, alpha))


@dataclass
class scale_color_d3(scale_discrete):
    """
    D3.js color scale.

    Args:
        palette: Palette name: "category10", "category20", "category20b", or "category20c".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "category10"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_d3(palette, alpha))


@dataclass
class scale_fill_d3(scale_discrete):
    """
    D3.js fill scale.

    Args:
        palette: Palette name: "category10", "category20", "category20b", or "category20c".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "category10"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_d3(palette, alpha))


@dataclass
class scale_color_observable(scale_discrete):
    """
    Observable color scale.

    Args:
        palette: Palette name. Currently only "observable10" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "observable10"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_observable(palette, alpha))


@dataclass
class scale_fill_observable(scale_discrete):
    """
    Observable fill scale.

    Args:
        palette: Palette name. Currently only "observable10" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "observable10"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_observable(palette, alpha))


@dataclass
class scale_color_primer(scale_discrete):
    """
    Primer design system color scale.

    Args:
        palette: Palette name. Currently only "mark17" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "mark17"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_primer(palette, alpha))


@dataclass
class scale_fill_primer(scale_discrete):
    """
    Primer design system fill scale.

    Args:
        palette: Palette name. Currently only "mark17" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "mark17"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_primer(palette, alpha))


@dataclass
class scale_color_atlassian(scale_discrete):
    """
    Atlassian design system color scale.

    Args:
        palette: Palette name. Currently only "categorical8" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "categorical8"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_atlassian(palette, alpha))


@dataclass
class scale_fill_atlassian(scale_discrete):
    """
    Atlassian design system fill scale.

    Args:
        palette: Palette name. Currently only "categorical8" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "categorical8"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_atlassian(palette, alpha))


@dataclass
class scale_color_iterm(scale_discrete):
    """
    iTerm color scale.

    Args:
        palette: Palette name. See `ITERM_PALETTES` for available options.
        variant: Palette variant. Either "normal" or "bright".
        alpha: Transparency level, between 0 and 1.

    Details:
        Preview all iTerm palettes: <https://nanx.me/ggsci-iterm/>.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "Rose Pine"
    variant: InitVar[str] = "normal"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, variant: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_iterm(palette, variant, alpha))


@dataclass
class scale_fill_iterm(scale_discrete):
    """
    iTerm fill scale.

    Args:
        palette: Palette name. See `ITERM_PALETTES` for available options.
        variant: Palette variant. Either "normal" or "bright".
        alpha: Transparency level, between 0 and 1.

    Details:
        Preview all iTerm palettes: <https://nanx.me/ggsci-iterm/>.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "Rose Pine"
    variant: InitVar[str] = "normal"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, variant: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_iterm(palette, variant, alpha))


@dataclass
class scale_color_locuszoom(scale_discrete):
    """
    LocusZoom color scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_locuszoom(palette, alpha))


@dataclass
class scale_fill_locuszoom(scale_discrete):
    """
    LocusZoom fill scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_locuszoom(palette, alpha))


@dataclass
class scale_color_igv(scale_discrete):
    """
    IGV color scale.

    Args:
        palette: Palette name: "default" or "alternating".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_igv(palette, alpha))


@dataclass
class scale_fill_igv(scale_discrete):
    """
    IGV fill scale.

    Args:
        palette: Palette name: "default" or "alternating".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_igv(palette, alpha))


@dataclass
class scale_color_cosmic(scale_discrete):
    """
    COSMIC color scale.

    Args:
        palette: Palette name: "hallmarks_dark", "hallmarks_light", or "signature_substitutions".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "hallmarks_dark"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_cosmic(palette, alpha))


@dataclass
class scale_fill_cosmic(scale_discrete):
    """
    COSMIC fill scale.

    Args:
        palette: Palette name: "hallmarks_dark", "hallmarks_light", or "signature_substitutions".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "hallmarks_dark"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_cosmic(palette, alpha))


@dataclass
class scale_color_uchicago(scale_discrete):
    """
    University of Chicago color scale.

    Args:
        palette: Palette name: "default", "light", or "dark".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_uchicago(palette, alpha))


@dataclass
class scale_fill_uchicago(scale_discrete):
    """
    University of Chicago fill scale.

    Args:
        palette: Palette name: "default", "light", or "dark".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_uchicago(palette, alpha))


@dataclass
class scale_color_startrek(scale_discrete):
    """
    Star Trek color scale.

    Args:
        palette: Palette name. Currently only "uniform" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "uniform"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_startrek(palette, alpha))


@dataclass
class scale_fill_startrek(scale_discrete):
    """
    Star Trek fill scale.

    Args:
        palette: Palette name. Currently only "uniform" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "uniform"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_startrek(palette, alpha))


@dataclass
class scale_color_tron(scale_discrete):
    """
    Tron Legacy color scale.

    Args:
        palette: Palette name. Currently only "legacy" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "legacy"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_tron(palette, alpha))


@dataclass
class scale_fill_tron(scale_discrete):
    """
    Tron Legacy fill scale.

    Args:
        palette: Palette name. Currently only "legacy" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "legacy"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_tron(palette, alpha))


@dataclass
class scale_color_futurama(scale_discrete):
    """
    Futurama color scale.

    Args:
        palette: Palette name. Currently only "planetexpress" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "planetexpress"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_futurama(palette, alpha))


@dataclass
class scale_fill_futurama(scale_discrete):
    """
    Futurama fill scale.

    Args:
        palette: Palette name. Currently only "planetexpress" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "planetexpress"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_futurama(palette, alpha))


@dataclass
class scale_color_rickandmorty(scale_discrete):
    """
    Rick and Morty color scale.

    Args:
        palette: Palette name. Currently only "schwifty" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "schwifty"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_rickandmorty(palette, alpha))


@dataclass
class scale_fill_rickandmorty(scale_discrete):
    """
    Rick and Morty fill scale.

    Args:
        palette: Palette name. Currently only "schwifty" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "schwifty"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_rickandmorty(palette, alpha))


@dataclass
class scale_color_simpsons(scale_discrete):
    """
    The Simpsons color scale.

    Args:
        palette: Palette name. Currently only "springfield" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "springfield"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_simpsons(palette, alpha))


@dataclass
class scale_fill_simpsons(scale_discrete):
    """
    The Simpsons fill scale.

    Args:
        palette: Palette name. Currently only "springfield" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "springfield"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_simpsons(palette, alpha))


@dataclass
class scale_color_flatui(scale_discrete):
    """
    Flat UI color scale.

    Args:
        palette: Palette name: "default", "flattastic", or "aussie".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_flatui(palette, alpha))


@dataclass
class scale_fill_flatui(scale_discrete):
    """
    Flat UI fill scale.

    Args:
        palette: Palette name: "default", "flattastic", or "aussie".
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_flatui(palette, alpha))


@dataclass
class scale_color_frontiers(scale_discrete):
    """
    Frontiers journal color scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["color"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_frontiers(palette, alpha))


@dataclass
class scale_fill_frontiers(scale_discrete):
    """
    Frontiers journal fill scale.

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
    """

    _aesthetics = ["fill"]

    palette: InitVar[str] = "default"
    alpha: InitVar[float] = 1.0

    def __post_init__(self, palette: str, alpha: float) -> None:
        super().__post_init__()
        setattr(self, "palette", pal_frontiers(palette, alpha))


def scale_color_gsea(
    palette: str = "default",
    *,
    alpha: float = 1.0,
    reverse: bool = False,
    **kwargs: Any,
) -> scale_color_gradientn:
    """
    GSEA GenePattern color scale (continuous/diverging).

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.
        **kwargs: Additional keyword arguments forwarded to plotnine.scale_color_gradientn.
    """
    colors = pal_gsea(palette, n=512, alpha=alpha, reverse=reverse)
    return scale_color_gradientn(colors=colors, **kwargs)


def scale_fill_gsea(
    palette: str = "default",
    *,
    alpha: float = 1.0,
    reverse: bool = False,
    **kwargs: Any,
) -> scale_fill_gradientn:
    """
    GSEA GenePattern fill scale (continuous/diverging).

    Args:
        palette: Palette name. Currently only "default" is available.
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.
        **kwargs: Additional keyword arguments forwarded to plotnine.scale_fill_gradientn.
    """
    colors = pal_gsea(palette, n=512, alpha=alpha, reverse=reverse)
    return scale_fill_gradientn(colors=colors, **kwargs)


def scale_color_bs5(
    palette: str = "blue", *, alpha: float = 1.0, reverse: bool = False, **kwargs: Any
) -> scale_color_gradientn:
    """
    Bootstrap 5 color scale (continuous/sequential).

    Args:
        palette: Palette name: "blue", "indigo", "purple", "pink", "red",
            "orange", "yellow", "green", "teal", "cyan", or "gray".
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.
        **kwargs: Additional keyword arguments forwarded to plotnine.scale_color_gradientn.
    """
    colors = pal_bs5(palette, n=512, alpha=alpha, reverse=reverse)
    return scale_color_gradientn(colors=colors, **kwargs)


def scale_fill_bs5(
    palette: str = "blue", *, alpha: float = 1.0, reverse: bool = False, **kwargs: Any
) -> scale_fill_gradientn:
    """
    Bootstrap 5 fill scale (continuous/sequential).

    Args:
        palette: Palette name: "blue", "indigo", "purple", "pink", "red",
            "orange", "yellow", "green", "teal", "cyan", or "gray".
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.
        **kwargs: Additional keyword arguments forwarded to plotnine.scale_fill_gradientn.
    """
    colors = pal_bs5(palette, n=512, alpha=alpha, reverse=reverse)
    return scale_fill_gradientn(colors=colors, **kwargs)


def scale_color_material(
    palette: str = "red", *, alpha: float = 1.0, reverse: bool = False, **kwargs: Any
) -> scale_color_gradientn:
    """
    Material Design color scale (continuous/sequential).

    Args:
        palette: Palette name: "red", "pink", "purple", "deep-purple", "indigo",
            "blue", "light-blue", "cyan", "teal", "green", "light-green",
            "lime", "yellow", "amber", "orange", "deep-orange", "brown",
            "grey", or "blue-grey".
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.
        **kwargs: Additional keyword arguments forwarded to plotnine.scale_color_gradientn.
    """
    colors = pal_material(palette, n=512, alpha=alpha, reverse=reverse)
    return scale_color_gradientn(colors=colors, **kwargs)


def scale_fill_material(
    palette: str = "red", *, alpha: float = 1.0, reverse: bool = False, **kwargs: Any
) -> scale_fill_gradientn:
    """
    Material Design fill scale (continuous/sequential).

    Args:
        palette: Palette name: "red", "pink", "purple", "deep-purple", "indigo",
            "blue", "light-blue", "cyan", "teal", "green", "light-green",
            "lime", "yellow", "amber", "orange", "deep-orange", "brown",
            "grey", or "blue-grey".
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.
        **kwargs: Additional keyword arguments forwarded to plotnine.scale_fill_gradientn.
    """
    colors = pal_material(palette, n=512, alpha=alpha, reverse=reverse)
    return scale_fill_gradientn(colors=colors, **kwargs)


def scale_color_tw3(
    palette: str = "blue", *, alpha: float = 1.0, reverse: bool = False, **kwargs: Any
) -> scale_color_gradientn:
    """
    Tailwind CSS 3 color scale (continuous/sequential).

    Args:
        palette: Palette name: "slate", "gray", "zinc", "neutral", "stone",
            "red", "orange", "amber", "yellow", "lime", "green", "emerald",
            "teal", "cyan", "sky", "blue", "indigo", "violet", "purple",
            "fuchsia", "pink", or "rose".
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.
        **kwargs: Additional keyword arguments forwarded to plotnine.scale_color_gradientn.
    """
    colors = pal_tw3(palette, n=512, alpha=alpha, reverse=reverse)
    return scale_color_gradientn(colors=colors, **kwargs)


def scale_fill_tw3(
    palette: str = "blue", *, alpha: float = 1.0, reverse: bool = False, **kwargs: Any
) -> scale_fill_gradientn:
    """
    Tailwind CSS 3 fill scale (continuous/sequential).

    Args:
        palette: Palette name: "slate", "gray", "zinc", "neutral", "stone",
            "red", "orange", "amber", "yellow", "lime", "green", "emerald",
            "teal", "cyan", "sky", "blue", "indigo", "violet", "purple",
            "fuchsia", "pink", or "rose".
        alpha: Transparency level, between 0 and 1.
        reverse: Whether to reverse the color order.
        **kwargs: Additional keyword arguments forwarded to plotnine.scale_fill_gradientn.
    """
    colors = pal_tw3(palette, n=512, alpha=alpha, reverse=reverse)
    return scale_fill_gradientn(colors=colors, **kwargs)


# Aliases for British spelling
scale_colour_npg = scale_color_npg
scale_colour_aaas = scale_color_aaas
scale_colour_nejm = scale_color_nejm
scale_colour_lancet = scale_color_lancet
scale_colour_jama = scale_color_jama
scale_colour_bmj = scale_color_bmj
scale_colour_jco = scale_color_jco
scale_colour_ucscgb = scale_color_ucscgb
scale_colour_d3 = scale_color_d3
scale_colour_observable = scale_color_observable
scale_colour_primer = scale_color_primer
scale_colour_atlassian = scale_color_atlassian
scale_colour_iterm = scale_color_iterm
scale_colour_locuszoom = scale_color_locuszoom
scale_colour_igv = scale_color_igv
scale_colour_cosmic = scale_color_cosmic
scale_colour_uchicago = scale_color_uchicago
scale_colour_startrek = scale_color_startrek
scale_colour_tron = scale_color_tron
scale_colour_futurama = scale_color_futurama
scale_colour_rickandmorty = scale_color_rickandmorty
scale_colour_simpsons = scale_color_simpsons
scale_colour_flatui = scale_color_flatui
scale_colour_frontiers = scale_color_frontiers
scale_colour_gsea = scale_color_gsea
scale_colour_bs5 = scale_color_bs5
scale_colour_material = scale_color_material
scale_colour_tw3 = scale_color_tw3



from collections.abc import Mapping, Sequence
from typing import cast

from plotnine.data import diamonds
from plotnine.geoms import geom_bar, geom_point, geom_smooth
from plotnine.ggplot import ggplot

from ggsci import example_barplot, example_scatterplot


def test_example_scatterplot_configuration() -> None:
    plot = example_scatterplot()

    assert isinstance(plot, ggplot)
    assert len(plot.layers) == 2
    assert plot.mapping["color"] == "cut"
    assert plot.mapping["x"] == "table"
    assert plot.mapping["y"] == "price"

    data = cast(Mapping[str, Sequence[float]], plot.data)
    carat = cast(Sequence[float], data["carat"])
    assert all(value >= 2.2 for value in carat)
    assert isinstance(plot.layers[0].geom, geom_point)
    assert isinstance(plot.layers[1].geom, geom_smooth)


def test_example_barplot_configuration() -> None:
    plot = example_barplot()

    assert isinstance(plot, ggplot)
    assert plot.data is diamonds
    assert plot.mapping["x"] == "color"
    assert plot.mapping["fill"] == "cut"
    assert len(plot.layers) == 1
    assert isinstance(plot.layers[0].geom, geom_bar)



"""
Example plots used in documentation and demos.
"""

from __future__ import annotations

from plotnine import (
    aes,
    element_blank,
    geom_bar,
    geom_point,
    geom_smooth,
    ggplot,
    position_dodge,
    theme,
    theme_minimal,
)
from plotnine.data import diamonds


def example_scatterplot() -> ggplot:
    """
    Create the scatter plot example used in documentation.

    Args:
        None.

    Returns:
        Plot configured for discrete color palette demonstrations.

    Raises:
        None.
    """
    large_diamonds = diamonds.loc[diamonds["carat"] >= 2.2]

    return (
        ggplot(large_diamonds, aes("table", "price", color="cut"))
        + geom_point(alpha=0.7)
        + geom_smooth(method="lm", alpha=0.05, size=1)
        + theme_minimal()
        + theme(
            axis_title_x=element_blank(),
            axis_title_y=element_blank(),
            panel_grid_minor=element_blank(),
        )
    )


def example_barplot() -> ggplot:
    """
    Create the bar plot example used in documentation.

    Args:
        None.

    Returns:
        Plot configured for discrete fill palette demonstrations.

    Raises:
        None.
    """
    return (
        ggplot(diamonds, aes("color", fill="cut"))
        + geom_bar(position=position_dodge())
        + theme_minimal()
        + theme(
            axis_title_x=element_blank(),
            axis_title_y=element_blank(),
            panel_grid_major_x=element_blank(),
            panel_grid_minor_y=element_blank(),
        )
    )



"""ASCII verifier.

Verifies whether all text files in the current directory contain only ASCII characters.
"""

import os
import sys
from typing import FrozenSet, List, Tuple


def is_text_file(path: str, n: int | None = None) -> bool:
    """
    Classify any file as text or binary.

    Algorithm adopted from "A Fast Method for Identifying Plain Text Files"
    in zlib (`doc/txtvsbin.txt`).

    Args:
        path: File path.
        n: Maximal number of bytes to read. Defaults to file size.

    Returns:
        True if the file is text, False if binary.
    """
    ALLOW: FrozenSet[int] = frozenset([9, 10, 13] + list(range(32, 256)))
    BLOCK: FrozenSet[int] = frozenset(list(range(0, 7)) + list(range(14, 32)))

    with open(path, "rb") as file:
        bytecode = bytes(file.read(n or os.path.getsize(path)))

    if not bytecode:
        return False

    cond1 = any(b in ALLOW for b in bytecode)
    cond2 = not any(b in BLOCK for b in bytecode)

    return cond1 and cond2


def find_non_ascii_lines(filepath: str) -> List[Tuple[int, str, List[int]]]:
    """
    Find lines containing non-ASCII characters in a text file.

    Args:
        filepath: Path to the text file.

    Returns:
        List of tuples (line_number, line_content, positions) where positions
        are the character indices of non-ASCII characters in the line.
    """
    non_ascii_lines = []

    try:
        with open(filepath, "r", encoding="utf-8") as f:
            for line_num, line in enumerate(f, 1):
                non_ascii_positions = []
                for pos, char in enumerate(line):
                    if ord(char) > 127:
                        non_ascii_positions.append(pos)

                if non_ascii_positions:
                    non_ascii_lines.append(
                        (line_num, line.rstrip("\n"), non_ascii_positions)
                    )
    except (UnicodeDecodeError, OSError) as e:
        # File might not be UTF-8 encoded or readable
        print(f"Warning: Can't read {filepath}: {e}")

    return non_ascii_lines


def highlight_non_ascii(line: str, positions: List[int]) -> str:
    """
    Highlight non-ASCII characters in a line.

    Args:
        line: The line content.
        positions: Positions of non-ASCII characters.

    Returns:
        Line with highlighted non-ASCII characters.
    """
    result = []
    for i, char in enumerate(line):
        if i in positions:
            # Show the character and its Unicode code point
            result.append(f"[{repr(char)[1:-1]}:U+{ord(char):04X}]")
        else:
            result.append(char)
    return "".join(result)


def scan_directory(root_dir: str) -> None:
    """
    Scan directory recursively for text files with non-ASCII characters.

    Args:
        root_dir: Root directory to scan.
    """
    # Directories to skip
    skip_dirs = {
        ".git",
        "__pycache__",
        "build",
        "dist",
        "site",
        "wheels",
        ".venv",
        ".pytest_cache",
        ".egg-info",
    }

    files_with_non_ascii = []
    total_files_scanned = 0

    for dirpath, dirnames, filenames in os.walk(root_dir):
        # Skip certain directories
        dirnames[:] = [
            d for d in dirnames if d not in skip_dirs and not d.endswith(".egg-info")
        ]

        for filename in filenames:
            filepath = os.path.join(dirpath, filename)

            # Check if it's a text file
            if not os.path.isfile(filepath):
                continue

            try:
                if is_text_file(filepath):
                    total_files_scanned += 1
                    non_ascii_lines = find_non_ascii_lines(filepath)

                    if non_ascii_lines:
                        # Use relative path for cleaner output
                        rel_path = os.path.relpath(filepath, root_dir)
                        files_with_non_ascii.append((rel_path, non_ascii_lines))
            except Exception as e:
                print(f"Warning: Could not process {filepath}: {e}")

    # Report results
    print("ASCII verification report")
    print("=" * 25)
    print(f"Scanned {total_files_scanned} text files in: {root_dir}")
    print()

    if not files_with_non_ascii:
        print("SUCCESS: All text files contain only ASCII characters!")
    else:
        print(f"FOUND: {len(files_with_non_ascii)} file(s) with non-ASCII characters:")
        print()

        for filepath, non_ascii_lines in files_with_non_ascii:
            print(f"File: {filepath}")
            print(f"  Lines with non-ASCII characters: {len(non_ascii_lines)}")

            # Show up to 5 example lines
            for line_num, line, positions in non_ascii_lines[:5]:
                print(
                    f"  Line {line_num}: {highlight_non_ascii(line[:100], positions)}"
                )
                if len(line) > 100:
                    print(f"           ... (truncated, {len(line)} chars total)")

            if len(non_ascii_lines) > 5:
                print(f"  ... and {len(non_ascii_lines) - 5} more line(s)")
            print()

        # Summary
        print("Summary:")
        print(f"  Files with non-ASCII: {len(files_with_non_ascii)}")
        total_lines = sum(len(lines) for _, lines in files_with_non_ascii)
        print(f"  Total lines affected: {total_lines}")


def main():
    # Use current directory if no argument provided
    if len(sys.argv) > 1:
        root_dir = sys.argv[1]
    else:
        root_dir = os.getcwd()

    if not os.path.isdir(root_dir):
        print(f"Error: {root_dir} is not a directory")
        sys.exit(1)

    scan_directory(root_dir)


if __name__ == "__main__":
    main()
